/* 
   nmidi - ALSA/MWPP MIDI client

   Copyright (C) 2003 Hans Fugal

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  

#include <termios.h>
#include <grp.h>
#include <pwd.h>
*/

#include <stdio.h>
#include <sys/types.h>
#include <getopt.h>
#include "system.h"
#include <alsa/asoundlib.h>
#include <pthread.h>
#include <assert.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <errno.h>

#include "nmidi.h"

#define EXIT_FAILURE 1

char *xmalloc ();
char *xrealloc ();
char *xstrdup ();

int main (int argc, char **argv)
{
	int i;
	int sockfd;
	struct sockaddr_in serv_addr;
	struct sockaddr_in my_addr,client_addr;
	int addr_size;
	int client_sockfd;
	struct hostent *hp;
	pthread_t out,in;
	struct timeval tv;

	program_name = argv[0];

	port = DEFAULT_PORT;
	hostname = "localhost";
	i = decode_switches (argc, argv);
	if (i < argc)
	{
		if (i < argc-1)
			usage(EXIT_FAILURE);
		else
		{
			hostname = strdup(argv[i]);
			client = 1;
		}
	}
	/* explicit subscription to ports */
	/* phew this is complicated. later!
	st = midi_status_new(1);
	if (src_client >= 0 && src_client != SND_SEQ_ADDRESS_SUBSCRIBERS)
		port_connect_from(st->ports[0].port, src_client, src_port);
	if (dest_client >= 0 && dest_client != SND_SEQ_ADDRESS_SUBSCRIBERS) {
		port_connect_to(st->ports[0].port, dest_client, dest_port);
		reset_all(st);
	}
	*/


	gettimeofday(&tv,NULL);
	basetime = (double)tv.tv_sec + ((double)tv.tv_usec * 1.0e-6);

	seq_handle = open_seq();
	if (client)
	{
		printf("Connecting to %s on port %d...\n", hostname, port);
		sockfd = socket(PF_INET,SOCK_STREAM,IPPROTO_TCP);

		if (sockfd < 0)
		{
			perror("socket");
			exit(EXIT_FAILURE);
		}

		serv_addr.sin_family = AF_INET;
		serv_addr.sin_port = htons(port);
		if ((hp = gethostbyname(hostname)) == NULL)
		{
			herror("gethostbyname");
			exit(EXIT_FAILURE);
		}
		serv_addr.sin_addr = *((struct in_addr *)hp->h_addr);
		memset(&(serv_addr.sin_zero), '\0', 8);
		if (connect(sockfd, (struct sockaddr*)&serv_addr, sizeof(serv_addr)) 
				< 0)
		{
			perror("connect");
			exit(EXIT_FAILURE);
		}
		printf("Connection established.\n");

	} else { /* server */
		sockfd = socket(PF_INET,SOCK_STREAM,IPPROTO_TCP);

		if (sockfd < 0)
		{
			perror("socket");
			exit(EXIT_FAILURE);
		}

		my_addr.sin_family = AF_INET;
		my_addr.sin_port = htons(port);
		my_addr.sin_addr.s_addr = htonl(INADDR_ANY);
		if (bind(sockfd, (struct sockaddr*)&my_addr, sizeof(my_addr)) < 0)
		{
			perror("bind");
			exit(EXIT_FAILURE);
		}

		printf("Listening on port %d...\n",port);
		if (listen(sockfd, 1) < 0)
		{
			perror("listen");
			exit(EXIT_FAILURE);
		}

		addr_size = sizeof(client_addr);
		client_sockfd = accept(sockfd, NULL, NULL);
		if (client_sockfd < 0)
		{
			perror("accept");
			exit(EXIT_FAILURE);
		}
		close(sockfd);
		sockfd = client_sockfd;
		printf ("Connection established.\n");

	}

	pthread_create(&in, NULL, inbound_thread, &sockfd);
	pthread_create(&out, NULL, outbound_thread, &sockfd);
	pthread_join(in, NULL);
	pthread_join(out, NULL);
	exit (0);
}

/* Set all the option flags according to the switches specified.
   Return the index of the first non-option argument.  */

static int decode_switches (int argc, char **argv)
{
	int c;

	while ((c = getopt(argc, argv, "hVp:d:s:")) != -1)
	{
		switch (c)
		{
			case 'h':
				usage (0);
				break;

			case 'V':
				printf ("nmidi %s\n", VERSION);
				exit (0);
				break;

			case 'p':
				port = atoi(optarg);
				break;

			case 's':
				if (parse_addr(optarg, &src_client, &src_port) < 0)
				{
					fprintf(stderr, "invalid argument %s for -s\n", optarg);
					exit(EXIT_FAILURE);
				}
				fprintf(stderr, "-s: Not yet implemented. Use aconnect.\n");
				/* usage(EXIT_FAILURE);*/
				break;

			case 'd':
				if (parse_addr(optarg, &dest_client, &dest_port) < 0)
				{
					fprintf(stderr, "invalid argument %s for -d\n", optarg);
					exit(EXIT_FAILURE);
				}
				fprintf(stderr, "-d: Not yet implemented. Use aconnect.\n");
				/* usage(EXIT_FAILURE);*/
				break;

			case 'u':
				use_udp = 1;
				break;
			default:
				usage (EXIT_FAILURE);
		}
	}

	return optind;
}


static void usage (int status)
{
	printf (_("%s - ALSA/MWPP MIDI client\n"), program_name);
	printf (_("\
usage: \n\
  client: %s [options] <hostname>\n\
  server: %s [options]\n"), program_name,program_name);
	printf (_("options:\n\
  -h               display this help and exit\n\
  -V               version information and exit\n\
  -p <port>        port to use (default is %d)\n\
  -d <client:port> destination ALSA client and port\n\
  -s <client:port> source ALSA client and port\n\
"),DEFAULT_PORT);
	exit (status);
}

snd_seq_t *open_seq() 
{
	snd_seq_t *seq_handle;
	char *client_name = "nmidi";

	if (snd_seq_open(&seq_handle, "default", SND_SEQ_OPEN_DUPLEX, 0) < 0) 
	{
		fprintf(stderr, "Error opening ALSA sequencer.\n");
		exit(1);
	}
	if (client)
		client_name = "nmidi_c";
	snd_seq_set_client_name(seq_handle, client_name);
	if ((portid = snd_seq_create_simple_port(seq_handle, client_name,
					SND_SEQ_PORT_CAP_WRITE|SND_SEQ_PORT_CAP_SUBS_WRITE|
					SND_SEQ_PORT_CAP_READ|SND_SEQ_PORT_CAP_SUBS_READ,
					SND_SEQ_PORT_TYPE_APPLICATION)) < 0) 
	{
		fprintf(stderr, "Error creating sequencer port.\n");
		exit(1);
	}
	return seq_handle;
}

void *outbound_thread(void*arg)
{
	int sockfd = *(int*)arg;
	int npfd;
	struct pollfd *pfd;
	unsigned char send_buf[MAX_PACKET_LEN];
	snd_seq_event_t ev_list[MAX_EVENTS];
	snd_seq_event_t *ev;
	int send_len;
	int i, sent;

	npfd = snd_seq_poll_descriptors_count(seq_handle, POLLIN);
	pfd = (struct pollfd *)alloca(npfd * sizeof(struct pollfd));
	snd_seq_poll_descriptors(seq_handle, pfd, npfd, POLLIN);
	while (1) 
	{
		if (poll(pfd, npfd, -1) > 0) 
		{
			i = 0;
			do {
				snd_seq_event_input(seq_handle, &ev);
				ev_list[i++] = *ev;
				snd_seq_free_event(ev);
			} while (snd_seq_event_input_pending(seq_handle,0) > 0);
			send_len = mwpp_encode(ev_list, i, send_buf);
#ifdef DEBUG
			/*
			printf("%d events in %d bytes",i,send_len);			
			print_chars("",send_buf,send_len);
			*/
#endif
			sent = 0;
			while (sent < send_len)
				sent += send(sockfd, send_buf + sent, send_len - sent, 0);
		}  
	}
}
void *inbound_thread(void*arg)
{
	struct pollfd pfd;
	int sockfd = *(int*)arg;
	unsigned char buf[MAX_PACKET_LEN];
	snd_seq_event_t ev_list[MAX_EVENTS];
	snd_seq_event_t *ev;
	int n_events;
	int len;
	int i;

	pfd.fd = sockfd;
	pfd.events = POLLIN | POLLHUP;
	
	while (poll(&pfd, 1, -1) > 0) 
	{
		len = recv(sockfd, buf, MAX_PACKET_LEN, 0);
		if (len == 0)
			break;
		if (len < 0)
		{
			fprintf(stderr,"recv error\n");
			sleep(10);
			continue;
		}
#ifdef DEBUG
		/*
		printf("packet of %d bytes",len);
		print_chars("",buf,len);
		*/
#endif
		/* pass to ALSA */
		n_events = mwpp_decode(buf, len, ev_list, MAX_EVENTS);
		for (i=0; i<n_events; i++)
		{
			ev = &ev_list[i];
			snd_seq_ev_set_subs(ev);
			snd_seq_ev_set_direct(ev);
			snd_seq_ev_set_source(ev, portid);
			snd_seq_event_output_direct(seq_handle, &ev_list[i]);
#ifdef DEBUG
			if (snd_seq_ev_is_note_type(ev))
			{
				struct timeval t;
				gettimeofday(&t, NULL);
				printf("%.6lf %d %d %d\n",
						(double)t.tv_sec + ((double)t.tv_usec * 1.0e-6),
						ev->data.note.channel,
						ev->data.note.note,
						ev->data.note.velocity);
			}
#endif
		}
	}
}
	

int mwpp_encode(snd_seq_event_t *events, int count, unsigned char *buf)
{
	snd_midi_event_t *dev;
	int bufsize;
	snd_seq_event_t *ev;
	int len;
	int i;
	unsigned char *bufp;
	int rlen;
	struct timeval tv;
	unsigned long now, dtime;
	
	len = 0;
	bufsize = 64; /* good size? */
	snd_midi_event_new(bufsize,&dev);
	snd_midi_event_init(dev);
	snd_midi_event_no_status(dev,1);

	bufp = buf;

	gettimeofday(&tv, NULL);
	now = ((double)tv.tv_sec + ((double)tv.tv_usec * 1.0e-6)) * SRATE;
	dtime = now - basetime;
	for (i=0; i<count; i++)
	{
		if (buf+MAX_PACKET_LEN - bufp < bufsize)
		{
			fprintf(stderr,"encoding buffer overflow\n");
			return -1;
		}
		ev = &events[i];
#ifdef DEBUG
			if (snd_seq_ev_is_note_type(ev))
			{
				struct timeval t;
				gettimeofday(&t, NULL);
				printf("%.6lf %d %d %d\n",
						(double)t.tv_sec + ((double)t.tv_usec * 1.0e-6),
						ev->data.note.channel,
						ev->data.note.note,
						ev->data.note.velocity);
			}
#endif
		if (snd_seq_ev_is_subscribe_type(ev))
		{
#ifdef DEBUG
			fprintf(stderr,"subscribe_type\n");
#endif
			continue;
		}
		/*
		memcpy(bufp,&dtime,4);
		bufp += 4;
		len += 4;
		*/
		rlen = snd_midi_event_decode(dev, bufp, bufsize, ev);  	
		if (rlen < 0)
		{
#ifdef DEBUG
			fprintf(stderr,"Error decoding event\n");
#endif
			return len;
		}
		len += rlen;
		bufp += rlen;
	}
	return len;
}
int mwpp_decode(unsigned char *buf, int len, 
		snd_seq_event_t *events, int count)
{
	snd_midi_event_t *dev;
	snd_seq_event_t *ev;
	int i=0;
	int n_ev=0;
	int rlen;
	unsigned long dtime;
	
	snd_midi_event_new(len,&dev);
	snd_midi_event_init(dev);


	while (i<len)
	{
		ev = &events[n_ev];
		/*
		memcpy(&dtime,buf+i,4);
		i+=4;
		*/
		rlen = snd_midi_event_encode(dev, buf+i, len-i, ev);
		if (rlen < 0)
		{
			fprintf(stderr,"encoding error\n");
			break;
		}
		if (rlen == SND_SEQ_EVENT_NONE)
		{
			fprintf(stderr,"unexpected end of stream while encoding\n");
			break;
		}
		i += rlen;
		n_ev++;
	}
	return n_ev;
}

/* "parse" one MIDI event. All it really does is say how many bytes the first
 * event is, beginning at byte 0 of buf returns < 0 on error */
int parse_event(unsigned char *buf, unsigned char *newbuf)
{
	int rval;
	int p1 = 0;
	int p2 = p1;
	unsigned char *p = buf;

	if (buf[p1] & 0x80)
	{
		newbuf[p2++] = buf[p1++];
		if (buf[p1] == 0xf0)
		{ /* sysex */
			while (buf[p1] == 0xf0 || !(buf[p1] & 0x80))
				newbuf[p2++] = buf[p1++];
			if (buf[p1] == 0xf7)
				newbuf[p2++] = buf[p1++];
		} 
		else if (buf[p1] & 0xf8 == 0xf8)
		{ /* system realtime messages */
			newbuf[p2++] = buf[p1++];
			/* return 1; */
		} 
		else
		{ /* other system and voice messages */
			while (!(buf[p1] & 0x80) || buf[p1] & 0xf8 == 0xf8)
			{
				if (buf[p1] & 0xf8 == 0xf8)
					p1++; /* ignore system realtime messages for now */
				else
					newbuf[p2++] = buf[p1++];
			}
		}
		rval = p2;
	} else {
		rval = -1;
	}
	return rval;
}

void print_chars(char *head, unsigned char *buff, int length)
{
	int i;
	int j;

	printf("%s\n",head);
	for(i = 0; i < length; i+=16) 
	{
		for (j=0; j< 16 ; j++)
		{
			if(j==8) printf(" ");
			if(i+j < length) 
				printf("%02x ",buff[i+j]);
			else
				printf("   ");
		}
		printf("  ");
		for (j=0; j< 16 ; j++)
		{
			if(j==8) printf(" ");
			if(i+j < length) 
			{
				if(isprint(buff[i+j]))
					printf("%c",buff[i+j]);
				else
					printf(".");
			}
		}
		printf("\n");

	}
}

/*
 * parse client:port address from command line
 * thanks to Takashi Iwai (aseqview)
 */
static int
parse_addr(char *arg, int *clientp, int *portp)
{
	char *p;
	if (isdigit(*arg)) {
		if ((p = strpbrk(arg, ":.")) == NULL)
			return -1;
		*clientp = atoi(arg);
		*portp = atoi(p + 1);
	} else {
		if (*arg == 's' || *arg == 'S') {
			*clientp = SND_SEQ_ADDRESS_SUBSCRIBERS;
			*portp = 0;
		} else
			return -1;
	}
	return 0;
}
/*
 * reset all stuff
 * thanks to Takashi Iwai (aseqview)
 */
/*
static void
reset_all(midi_status_t *st)
{
	int i, p;

	st->midi_mode = MIDI_MODE_GM;
	st->timer_update = TRUE;
	display_midi_mode(st);
	for (p = 0; p < st->num_ports; p++) {
		port_status_t *port = &st->ports[p];
		if (port->index < 0)
			continue;
		for (i = 0; i < MIDI_CHANNELS; i++) {
			channel_status_t *chst = &port->ch[i];
			all_sounds_off(chst);
			reset_controllers(chst);
			change_pitch(port, i, 0);
			change_program(port, i, 0);
			if (is_redirect(port))
				send_resets(chst);
			if (i == 9)
				chst->is_drum = 1;
			else
				chst->is_drum = 0;
			set_vel_bar_color(chst->w_vel, chst->is_drum);
		}
		if (is_redirect(port))
			port_flush_event(port->port);
	}
}
*/
